@@ -14,8 +14,8 @@ from pymodelextractor.learners.observation_table_learners.observation_table impo
 from pythautomata.automata.wheighted_automaton_definition.weighted_state import WeightedState
 
 class BoundedPDFAQuantizationNAryTreeLearner(PDFAQuantizationNAryTreeLearner):
-    def __init__(self, partitioner, max_states, max_query_length, max_seconds_run=None, generate_partial_hipothesis = False):
-        super().__init__(partitioner)
+    def __init__(self, partitioner, max_states, max_query_length, max_seconds_run=None, generate_partial_hipothesis = False, pre_cache_queries_for_building_hipothesis = False):
+        super().__init__(partitioner, pre_cache_queries_for_building_hipothesis)
         self._max_states = max_states
         self._max_query_length = max_query_length
         self._max_seconds_run = max_seconds_run
@@ -50,7 +50,7 @@ class BoundedPDFAQuantizationNAryTreeLearner(PDFAQuantizationNAryTreeLearner):
             self._exceeded_max_states = True
         except QueryLengthExceededException:
             print("QueryLengthExceeded")
-            self._exceeded_max_mq_length = True
+            self._exceeded_max_mq_length = True        
 
         if not self._exceeded_max_states and self._generate_partial_hipothesis and (self._exceeded_max_mq_length or self._exceded_time_bound) and len(self._tree.leaves)>0:         
             partial_hipothesis = self.partial_hipothesis()
@@ -74,11 +74,12 @@ class BoundedPDFAQuantizationNAryTreeLearner(PDFAQuantizationNAryTreeLearner):
             self._exceeded_max_mq_length = True
             return True, None
 
-    def partial_hipothesis(self) -> PDFA:
+    def partial_hipothesis(self) -> PDFA:        
+        if self._pre_cache_queries_for_building_hipothesis:
+                self._tree.cache_queries_for_building_hipothesis()        
         states = {}
         symbols = list(self._alphabet.symbols)
         symbols.sort()
-        
         for leaf_str, leaf in self._tree.leaves.items():
             initial_weight = 1 if leaf_str == epsilon else 0
             terminal_symbol_probability = leaf.probabilities[self.terminal_symbol]
@@ -86,17 +87,19 @@ class BoundedPDFAQuantizationNAryTreeLearner(PDFAQuantizationNAryTreeLearner):
             states[leaf_str] = state
         unknown_state =  WeightedState(self._tree.unknown_leaf, 0, 0)     
         states[self._tree.unknown_leaf] = unknown_state
-        unknown_state_is_accessed = False
+        accessed_states = set()
         for access_string, state in states.items():
             if access_string!=self._tree.unknown_leaf:
                 for symbol in symbols:
                     access_string_of_transition, _ = self._tree.sift(access_string + symbol, update=False)
-                    if access_string_of_transition == self._tree.unknown_leaf:
-                        unknown_state_is_accessed = True
+                    if access_string_of_transition != access_string:
+                        accessed_states.add(access_string_of_transition)
                     state.add_transition(symbol, states[access_string_of_transition],
-                                            self._tree.leaves[access_string].probabilities[symbol])
-        if not unknown_state_is_accessed:
-            del states[self._tree.unknown_leaf]
+                                            self._tree.leaves[access_string].probabilities[symbol])        
+        
+        for state in list(states.keys()).copy():
+            if state not in accessed_states and states[state].initial_weight != 1:
+                del states[state]
 
         comparator = WFAToleranceComparator()
         states = set(states.values())
